# PRIOR INFORMATION TRANSFER FOR GRAPHICAL MODELS OF SEWELL AND SHAH'S DATA.
#
# Written by Radford M. Neal, July 2001.
#
# These functions were used for Example 2 in my paper on ``Transferring
# prior information between models using imaginary data'', which is about
# directed graphical models for the data analysed in the following papers:
#
#  Sewell, W. H. and Shah, V. P. (1968) ``Social class, parental
#    encouragement, and educational aspirations'', American Journal
#    of Sociology, vol. 73, pp. 559-572.
#
#  Heckerman, D., Meek, C., and Cooper, G. F. (1999) ``A Bayesian
#    approach to causal discovery'', in C. Glymour and G. F. Cooper
#    (editors), Computation, Causation, and Discovery, AAAI Press.
#
# These functions use the functions defined in log-exp.r.  They assume
# that the variables e2.keep.cp, e2.tbl, and e2.alpha have been set to 
# the flag saying whether to keep the CP variable, to the table of
# counts in the data, and to the table of base Dirichlet parameters.


# FUNCTION TO COMPUTE A BIT OF THE LOG MARGINAL LIKELIHOOD, FOR DIRICHLET PRIOR.
#
# Compute the marginal likelihood for the part of the model corresponding to
# one node in the graph.  The a and tbl arguments are arrays with dimensions
# for the node and its parents, with the node itself being indexed by i.  The
# a array contains the Dirichlet prior parameters; the tbl array the counts
# from the data.

graph.ml.dirichlet <- function (i, a, tbl)
{
  if (is.null(dim(a)))
  { if (i!=1) stop("i!=1 when a is a vector")
    ai <- sum(a)
  }
  else
  { ai <- apply(a,(1:length(dim(a)))[-i],sum)
  }

  if (is.null(dim(tbl)))
  { if (i!=1) stop("i!=1 when tbl is a vector")
    tbli <- sum(tbl)
  }
  else
  { tbli <- apply(tbl,(1:length(dim(a)))[-i],sum)
  }

  sum(lgamma(a+tbl)-lgamma(a)) - sum(lgamma(ai+tbli)-lgamma(ai))
}


# COMPUTE LOG MARGINAL LIKELIHOOD FOR MODEL A, WITH GIVEN DIRICHLET PRIOR.

graph.ml.a <- function (alpha, tbl)
{
  # sex 
  ml <- graph.ml.dirichlet (1, apply(alpha,c(1),sum),       
                               apply(tbl,c(1),sum))

  # iq | ses
  ml <- ml + graph.ml.dirichlet (2, apply(alpha,c(2,3),sum),
                                    apply(tbl,c(2,3),sum))

  # pe | sex, iq, ses
  ml <- ml + graph.ml.dirichlet (4, apply(alpha,c(1,2,3,4),sum),
                                    apply(tbl,c(1,2,3,4),sum)) 

  # ses
  ml <- ml + graph.ml.dirichlet (1, apply(alpha,c(2),sum),
                                    apply(tbl,c(2),sum))

  # cp | iq, pe, ses
  if (e2.keep.cp)
  { ml <- ml + graph.ml.dirichlet (4, apply(alpha,c(2,3,4,5),sum),
                                      apply(tbl,c(2,3,4,5),sum))
  }

  ml
}


# COMPUTE LOG MARGINAL LIKELIHOOD FOR MODEL B, WITH GIVEN DIRICHLET PRIOR.

graph.ml.b <- function (alpha, tbl)
{
  # sex 
  ml <- graph.ml.dirichlet (1, apply(alpha,c(1),sum),
                               apply(tbl,c(1),sum))

  # iq | pe, ses
  ml <- ml + graph.ml.dirichlet (2, apply(alpha,c(2,3,4),sum),
                                    apply(tbl,c(2,3,4),sum))

  # pe | sex, ses
  ml <- ml + graph.ml.dirichlet (3, apply(alpha,c(1,2,4),sum),
                                    apply(tbl,c(1,2,4),sum))

  # ses
  ml <- ml + graph.ml.dirichlet (1, apply(alpha,c(2),sum),
                                    apply(tbl,c(2),sum))

  # cp | iq, pe, ses
  if (e2.keep.cp)
  { ml <- ml + graph.ml.dirichlet (4, apply(alpha,c(2,3,4,5),sum),
                                      apply(tbl,c(2,3,4,5),sum))
  }

  ml
}


# GENERATE PROBABILITIES FROM MODEL A, WITH GIVEN DIRICHLET PRIOR.

graph.gen.a <- function (alpha)
{
  # Generate conditional probabilities from the prior.

  p.sex <- rgamma(2,apply(alpha,c(1),sum))
  p.sex <- p.sex/sum(p.sex)

  p.ses <- rgamma(4,apply(alpha,c(2),sum))
  p.ses <- p.ses/sum(p.ses)

  p.pe <- rgamma(2*4*4*2,apply(alpha,c(1,2,3,4),sum))
  dim(p.pe) <- c(2,4,4,2)
  for (sex in 1:2) 
  { for (ses in 1:4) 
    { for (iq in 1:4)
      { p.pe[sex,ses,iq,] <- p.pe[sex,ses,iq,]/sum(p.pe[sex,ses,iq,])
      }
    }
  }

  p.iq <- rgamma(4*4,apply(alpha,c(2,3),sum))
  dim(p.iq) <- c(4,4)
  for (ses in 1:4)
  { p.iq[ses,] <- p.iq[ses,]/sum(p.iq[ses,])
  }

  if (e2.keep.cp)
  { p.cp <- rgamma(4*4*2*2,apply(alpha,c(2,3,4,5),sum))
    dim(p.cp) <- c(4,4,2,2)
    for (ses in 1:4)
    { for (iq in 1:4)
      { for (pe in 1:2)
        { p.cp[ses,iq,pe,] <- p.cp[ses,iq,pe,]/sum(p.cp[ses,iq,pe,])
        }
      }  
    }
  }

  # Find probabilities for all combinations of values using these parameters.

  graph.comb.pr.a (p.sex, p.ses, p.pe, p.iq, p.cp)
}


# GENERATE PARTS OF MODEL A, WITH GIVEN DIRICHLET PRIOR.  Returns a 
# list with components sex, ses, pe, iq, and cp (if not suppressed).
# Thse are each generated from the prior if they are NULL in the arguments
# passed.  If they are non-null, but the index (from 1) of the part is
# given as the update argument, one of the conditional distributions for
# that part is selected at random, and changed to one drawn from the prior.

graph.gen.part.a <- function (alpha, update=0,
  p.sex=NULL, p.ses=NULL, p.pe=NULL, p.iq=NULL, p.cp=NULL)
{
  if (is.null(p.sex) || update==1)
  { p.sex <- rgamma(2,apply(alpha,c(1),sum))
    p.sex <- p.sex/sum(p.sex)
  }

  if (is.null(p.ses) || update==2)
  { p.ses <- rgamma(4,apply(alpha,c(2),sum))
    p.ses <- p.ses/sum(p.ses)
  }

  if (is.null(p.pe))
  { p.pe <- rgamma(2*4*4*2,apply(alpha,c(1,2,3,4),sum))
    dim(p.pe) <- c(2,4,4,2)
    for (sex in 1:2) 
    { for (ses in 1:4) 
      { for (iq in 1:4)
        { p.pe[sex,ses,iq,] <- p.pe[sex,ses,iq,]/sum(p.pe[sex,ses,iq,])
        }
      }
    }
  }
  else if (update==3)
  { sex <- sample(2,1)
    ses <- sample(4,1)
    iq  <- sample(4,1)
    p.pe[sex,ses,iq,] <- rgamma(2,apply(alpha,c(1,2,3,4),sum)[sex,ses,iq,])
    p.pe[sex,ses,iq,] <- p.pe[sex,ses,iq,]/sum(p.pe[sex,ses,iq,])
  }

  if (is.null(p.iq))
  { p.iq <- rgamma(4*4,apply(alpha,c(2,3),sum))
    dim(p.iq) <- c(4,4)
    for (ses in 1:4)
    { p.iq[ses,] <- p.iq[ses,]/sum(p.iq[ses,])
    }
  }
  else if (update==4)
  { ses <- sample(4,1)
    p.iq[ses,] <- rgamma(4,apply(alpha,c(2,3),sum)[ses,])
    p.iq[ses,] <- p.iq[ses,]/sum(p.iq[ses,])
  }

  if (is.null(p.cp) && e2.keep.cp)
  { p.cp <- rgamma(4*4*2*2,apply(alpha,c(2,3,4,5),sum))
    dim(p.cp) <- c(4,4,2,2)
    for (ses in 1:4)
    { for (iq in 1:4)
      { for (pe in 1:2)
        { p.cp[ses,iq,pe,] <- p.cp[ses,iq,pe,]/sum(p.cp[ses,iq,pe,])
        }
      }  
    }
  }
  else if (update==5 && e2.keep.cp)
  { ses <- sample(4,1)
    iq  <- sample(4,1)
    pe  <- sample(2,1)
    p.cp[ses,iq,pe,] <- rgamma(2,apply(alpha,c(2,3,4,5),sum)[ses,iq,pe,])
    p.cp[ses,iq,pe,] <- p.cp[ses,iq,pe,]/sum(p.cp[ses,iq,pe,])
  }

  list (sex=p.sex, ses=p.ses, pe=p.pe, iq=p.iq, cp=p.cp)
}


# COMBINE PROBABILITIES FOR MODEL A.

graph.comb.pr.a <- function (p.sex, p.ses, p.pe, p.iq, p.cp)
{
  p <- rep(0,e2.nvals) 
  dim(p) <- dim(e2.tbl)

  if (e2.keep.cp)
  { for (sex in 1:2)
    { for (ses in 1:4)
      { for (iq in 1:4)
        { for (pe in 1:2)
          { for (cp in 1:2)
            { p[sex,ses,iq,pe,cp] <- 
                p.sex[sex] * p.ses[ses] * p.pe[sex,ses,iq,pe] * p.iq[ses,iq] *
                p.cp[ses,iq,pe,cp]
            }
          }
        }
      }
    }
  }
  else
  { for (sex in 1:2)
    { for (ses in 1:4)
      { for (iq in 1:4)
        { for (pe in 1:2)
          { p[sex,ses,iq,pe] <- 
              p.sex[sex] * p.ses[ses] * p.pe[sex,ses,iq,pe] * p.iq[ses,iq]
          }
        }
      }
    }
  }

  p
}


# GENERATE PROBABILITIES FROM MODEL B, WITH GIVEN DIRICHLET PRIOR.

graph.gen.b <- function (alpha)
{
  # Generate conditional probabilities from the prior.

  p.sex <- rgamma(2,apply(alpha,c(1),sum))
  p.sex <- p.sex/sum(p.sex)

  p.ses <- rgamma(4,apply(alpha,c(2),sum))
  p.ses <- p.ses/sum(p.ses)

  p.pe <- rgamma(2*4*2,apply(alpha,c(1,2,4),sum))
  dim(p.pe) <- c(2,4,2)
  for (sex in 1:2) 
  { for (ses in 1:4) 
    { p.pe[sex,ses,] <- p.pe[sex,ses,]/sum(p.pe[sex,ses,])
    }
  }

  p.iq <- rgamma(4*4*2,apply(alpha,c(2,3,4),sum))
  dim(p.iq) <- c(4,4,2)
  for (ses in 1:4)
  { for (pe in 1:2)
    { p.iq[ses,,pe] <- p.iq[ses,,pe]/sum(p.iq[ses,,pe])
    }
  }

  if (e2.keep.cp)
  { p.cp <- rgamma(4*4*2*2,apply(alpha,c(2,3,4,5),sum))
    dim(p.cp) <- c(4,4,2,2)
    for (ses in 1:4)
    { for (iq in 1:4)
      { for (pe in 1:2)
        { p.cp[ses,iq,pe,] <- p.cp[ses,iq,pe,]/sum(p.cp[ses,iq,pe,])
        }
      }  
    }
  }

  # Find probabilities for all combinations of values using these parameters.

  graph.comb.pr.b (p.sex, p.ses, p.pe, p.iq, p.cp)
}


# GENERATE PARTS OF MODEL B, WITH GIVEN DIRICHLET PRIOR.  Returns a 
# list with components sex, ses, pe, iq, and cp (if not suppressed).
# Thse are each generated from the prior if they are NULL in the arguments
# passed.  If they are non-null, but the index (from 1) of the part is
# given as the update argument, one of the conditional distributions for
# that part is selected at random, and changed to one drawn from the prior.

graph.gen.part.b <- function (alpha, update=0,
  p.sex=NULL, p.ses=NULL, p.pe=NULL, p.iq=NULL, p.cp=NULL)
{
  if (is.null(p.sex) || update==1)
  { p.sex <- rgamma(2,apply(alpha,c(1),sum))
    p.sex <- p.sex/sum(p.sex)
  }

  if (is.null(p.ses) || update==2)
  { p.ses <- rgamma(4,apply(alpha,c(2),sum))
    p.ses <- p.ses/sum(p.ses)
  }

  if (is.null(p.pe))
  { p.pe <- rgamma(2*4*2,apply(alpha,c(1,2,4),sum))
    dim(p.pe) <- c(2,4,2)
    for (sex in 1:2) 
    { for (ses in 1:4) 
      { p.pe[sex,ses,] <- p.pe[sex,ses,]/sum(p.pe[sex,ses,])
      }
    }
  }
  else if (update==3)
  { sex <- sample(2,1)
    ses <- sample(4,1)
    p.pe[sex,ses,] <- rgamma(2,apply(alpha,c(1,2,4),sum)[sex,ses,])
    p.pe[sex,ses,] <- p.pe[sex,ses,]/sum(p.pe[sex,ses,])
  }

  if (is.null(p.iq))
  { p.iq <- rgamma(4*4*2,apply(alpha,c(2,3,4),sum))
    dim(p.iq) <- c(4,4,2)
    for (ses in 1:4)
    { for (pe in 1:2)
      { p.iq[ses,,pe] <- p.iq[ses,,pe]/sum(p.iq[ses,,pe])
      }
    }
  }
  else if (update==4)
  { ses <- sample(4,1)
    pe  <- sample(2,1)
    p.iq[ses,,pe] <- rgamma(4,apply(alpha,c(2,3,4),sum)[ses,,pe])
    p.iq[ses,,pe] <- p.iq[ses,,pe]/sum(p.iq[ses,,pe])
  }

  if (is.null(p.cp) && e2.keep.cp)
  { p.cp <- rgamma(4*4*2*2,apply(alpha,c(2,3,4,5),sum))
    dim(p.cp) <- c(4,4,2,2)
    for (ses in 1:4)
    { for (iq in 1:4)
      { for (pe in 1:2)
        { p.cp[ses,iq,pe,] <- p.cp[ses,iq,pe,]/sum(p.cp[ses,iq,pe,])
        }
      }  
    }
  }
  else if (update==5 && e2.keep.cp)
  { ses <- sample(4,1)
    iq  <- sample(4,1)
    pe  <- sample(2,1)
    p.cp[ses,iq,pe,] <- rgamma(2,apply(alpha,c(2,3,4,5),sum)[ses,iq,pe,])
    p.cp[ses,iq,pe,] <- p.cp[ses,iq,pe,]/sum(p.cp[ses,iq,pe,])
  }

  list (sex=p.sex, ses=p.ses, pe=p.pe, iq=p.iq, cp=p.cp)
}


# COMBINE PROBABILITIES FOR MODEL B.

graph.comb.pr.b <- function (p.sex, p.ses, p.pe, p.iq, p.cp)
{
  p <- rep(0,e2.nvals) 
  dim(p) <- dim(e2.tbl)

  if (e2.keep.cp)
  { for (sex in 1:2)
    { for (ses in 1:4)
      { for (iq in 1:4)
        { for (pe in 1:2)
          { for (cp in 1:2)
            { p[sex,ses,iq,pe,cp] <- 
                p.sex[sex] * p.ses[ses] * p.pe[sex,ses,pe] * p.iq[ses,iq,pe] *
                p.cp[ses,iq,pe,cp]
            }
          }
        }
      }
    }
  }
  else
  { for (sex in 1:2)
    { for (ses in 1:4)
      { for (iq in 1:4)
        { for (pe in 1:2)
          { p[sex,ses,iq,pe] <- 
              p.sex[sex] * p.ses[ses] * p.pe[sex,ses,pe] * p.iq[ses,iq,pe]
          }
        }
      }
    }
  }

  p
}


# COMPUTE LOG MARGINAL LIKELIHOOD WITH PRIOR INFORMATION TRANSFER, SIMPLE MC.
#
# The first two arguments are a procedure that generates probabilities
# from the donor model's prior and a procedure that calculates the marginal
# likelihood for the recipient model.  The third argument should be
# the type of weighting, either "donor" or "jointly".  Remaining arguments 
# give k, omega, and the Monte Carlo sample size, N.  The result is a list
# with the following elements:
#
#    ml     estimate of the log of the marginal likelihood for the entire data
#    se     standard error for the estimate of the log marginal likelihood
#    pn     log normalizing constant for the prior, 0 for donor-weighted prior
#    mlv    the N log values times weights used in finding the estimate
#    w      the N weights used in finding the estimate
#    cpu    Amount of user CPU time used for the computation

graph.pit.smc <- function (donor.gen, recipient.ml, wt, k=Inf, mass=k, N=100)
{
  stime <- proc.time()[1]

  if (mass==Inf)
  { stop("The mass argument must have a finite value")
  }

  if (wt!="donor" && wt!="jointly")
  { stop("Weight type (wt) must be either \"donor\" or \"jointly\"")
  }
  jointly.weighted <- wt=="jointly"

  mlv <- rep(0,N)     # log marginal likelihoods based on N imaginary data sets
  w   <- rep(0,N)     # logs of weights to be given to each value of y,
                      #   all zero for donor-weighted transfer
  for (i in 1:N)
  { 
    # Generate probabilities from donor model's prior.

    p <- donor.gen(e2.alpha)

    # Produce counts to add to pro forma prior.

    if (k==Inf)
    { cnts <- mass*p
    }
    else
    { cnts <- (mass/k) * 
                (table(c(1:e2.nvals,sample(1:e2.nvals,k,replace=T,prob=p))) - 1)
      dim(cnts) <- dim(e2.tbl)
    }

    # Find the log weight for this y when using jointly-weighted transfer.  
    # (The log weight is left at zero for donor-weighted transfer.)

    if (jointly.weighted)
    { w[i] <- recipient.ml (e2.alpha, cnts)
    }

    # Record the log marginal likelihood based on this imaginary data, with
    # weight added for jointly-weighted transfer.

    mlv[i] <- recipient.ml (e2.alpha+cnts, e2.tbl) + w[i]
  }

  # Compute and return the final estimate and its standard error.

  log.sum.exp.w <- log.sum.exp(w)
  ml <- log.sum.exp(mlv) - log.sum.exp.w

  se <- exp (log.sum.exp(log.sqdiff.exp(mlv,ml+w))/2 - log.sum.exp.w - ml)

  pn <- log.sum.exp.w - log(N)
 
  cpu <- proc.time()[1] - stime

  return (ml, se, pn, mlv, w, cpu)
}


# COMPUTE MARGINAL LIKELIHOOD WITH JOINTLY-WEIGHTED TRANSFER, BY ANN IMP SAMP.
#
# The first two arguments are the procedures that generate probabilities
# from the donor model's prior and recipient models' prior.  The remaining 
# arguments are as follows:
#
#    k          number of imaginary data points (default is infinite)
#    mass       mass of imaginary data (default k, required when k infinite)
#    N          number of annealing runs to do
#    alpha      Vector of powers to raise the likelihood to; 0 added at front
#    omega      Vector of powers to raise the imaginary likelihood to.  Must
#               start with 0 if the prior is jointly-weigthed; element equal to
#               mass is added to the end.  Defaults to nothing, which is 
#               appropriate for a donor-weighted prior
#
# The result is a list with the following elements:
#
#    ml     estimate of the log of the marginal likelihood for the entire data
#    se     standard error for the estimate of the log marginal likelihood
#    pn     log normalizing constant for the prior, 0 for donor-weighted prior
#    ess    effective sample size
#    cpu    Amount of user CPU time used for the computation
#
# The effective sample size is N divided by one plus the variance of the
# normalized importance weights.  It is a rough measure of the number of 
# unweighted data points that would be equivalent to the weighted sample.
#
# This version explicitly represents all of phi, y, and theta, and updates
# them by Gibbs sampling, except when k is infinite, when phi and y are
# updated by Metropolis-Hastings.

graph.jw.ais <- function (donor.gen, recipient.gen, 
                          k=Inf, mass=k, N=100, alpha, omega=numeric(0))
{
  stime <- proc.time()[1]

  # Check arguments for validity.

  if (length(omega)==0 || omega[1]!=0)
  { stop("Annealing schedule for omega must start with zero")
  }

  # MAIN LOOP.  Does one annealing run each time around, storing the logs of
  # the importance weights for the runs in w.  The importance weights at
  # the point where we reach the prior (omega equal to mass) are recorded
  # in w0 (this is needed only for jointly-weighted transfer).  

  w <- rep(0,N)
  w0 <- rep(0,N)

  omga <- c(omega,rep(mass,length(alpha)+2))
  alph <- c(rep(0,length(omega)+1),alpha,1)

  W <- matrix (0, N, length(omega)+length(alpha)+1)

  for (i in 1:N)
  {
    for (j in 1:(length(omega)+length(alpha)+1))
    {
      # Draw an independent state at the start of the run. 

      if (j==1)
      { p.d <- donor.gen(e2.alpha)
        if (k==Inf)
        { yt <- p.d
        }
        else
        { y <- sample (length(p.d), k, replace=T, prob=p.d)
          yt <- (table (c(y, 1:length(p.d))) - 1) / k
        }
        dim(yt) <- dim(e2.tbl)
        p.r <- recipient.gen(e2.alpha)
      }

      # Do a Gibbs sampling scan after the start of the run.

      if (j>1)
      {
        # Sample for theta (p.r) given data and y.

        p.r <- recipient.gen (e2.alpha + alph[j]*e2.tbl + omga[j]*yt)

        # Sample for y given phi (p.d) and theta (p.r). Skip when k is infinite.

        if (k!=Inf)
        { p <- p.d * p.r^(omga[j]/k)
          p <- p/sum(p)
          y <- sample(length(p), k, replace=T, prob=p)
          yt <- (table (c(y, 1:length(p))) - 1) / k
          dim(yt) <- dim(e2.tbl)
        }

        # Sample for phi (p.d) given y.  When k is infinite, sample for
        # phi and y simultaneously, with a Metropolis-Hastings update
        # tht proposes phi (and consequent y) from the donor's prior.

        if (k==Inf)
        { np.d <- donor.gen (e2.alpha)
          nyt <- np.d
          d <- sum(omga[j]*(yt-nyt)*log(p.r))
          if (runif(1)<exp(-d))
          { p.d <- np.d
            yt <- nyt
          }
        }
        else
        { p.d <- donor.gen (e2.alpha + yt)
        }
      }

      # Update the log of the weight for this run.

      w[i] <- w[i] + sum ((omga[j+1]*yt + alph[j+1]*e2.tbl) * log(p.r)) -
                     sum ((omga[j]*yt + alph[j]*e2.tbl) * log(p.r))

      # Record the weight at the point where the prior is reached.

      if (j==length(omega))
      { w0[i] <- w[i]
      }

      W[i,j] <- w[i]
    }

    cat("Annealing run ",i,", log weight ",round(w[i],2),"\n",sep="")
  }

  # Find final estimates.

  log.sum.exp.w  <- log.sum.exp(w)
  log.sum.exp.w0 <- log.sum.exp(w0)

  pn <- log.sum.exp.w0 - log(N)

  ml <- log.sum.exp.w - log(N)

  se <- exp (log.sum.exp(log.sqdiff.exp(w,ml))/2-log(N) - ml)

  ml <- ml - pn

  ess <- N / (1+var(exp(w-log.sum.exp.w+log(N))))

  cpu <- proc.time()[1] - stime

  return (ml, pn, se, ess, cpu, w0, w, W)
}


# COMPUTE MARGINAL LIKELIHOOD WITH DONOR-WEIGHTED TRANSFER, BY ANN IMP SAMP, V1.
#
# The first two arguments are the procedure that generate probabilities
# from the donor model's prior and the procedure that calculates the marginal 
# likelihood for the recipient model.  The remaining arguments are as follows:
#
#    k          number of imaginary data points (default is infinite)
#    mass       mass of imaginary data (default k, required if k infinite)
#    N          number of annealing runs to do
#    alpha      Vector of powers to raise the likelihood to; 0 is added at the
#               front and 1 is added at the back
#
# The result is a list with the following elements:
#
#    ml     estimate of the log of the marginal likelihood for the entire data
#    se     standard error for the estimate of the log marginal likelihood
#    ess    effective sample size
#    cpu    Amount of user CPU time used for the computation
#
# The effective sample size is N divided by one plus the variance of the
# normalized importance weights.  It is a rough measure of the number of 
# unweighted data points that would be equivalent to the weighted sample.
#
# This version represents only phi and y, marginalizing away theta, and updates
# phi and y by Metropolis-Hastings updates with proposals for phi from the 
# donor model's prior.

graph.dw.ais1 <- function (donor.gen, recipient.ml, k=Inf, mass=k, N=100, alpha)
{
  stime <- proc.time()[1]

  # MAIN LOOP.  Does one annealing run each time around, storing the logs of
  # the importance weights for the runs in w. 

  w <- rep(0,N)

  alph <- c(0,alpha,1)

  W <- matrix (0, N, length(alpha)+1)
  rej <- rep(0,length(alpha)+1)

  for (i in 1:N)
  {
    for (j in 1:(length(alpha)+1))
    {
      # Draw an independent state at the start of the run. 

      if (j==1)
      { phi <- donor.gen(e2.alpha)
        if (k==Inf)
        { yt <- phi
        }
        else
        { y <- sample (length(phi), k, replace=T, prob=phi)
          yt <- (table (c(y, 1:length(phi))) - 1) / k
        }
        dim(yt) <- dim(e2.tbl)
      }

      # Do a M-H update after the start of the run.

      if (j>1)
      {
        ml <- recipient.ml (e2.alpha+mass*yt, alph[j]*e2.tbl)

        nphi <- donor.gen(e2.alpha)
        if (k==Inf)
        { nyt <- nphi
        }
        else
        { ny <- sample (length(nphi), k, replace=T, prob=nphi)
          nyt <- (table (c(ny, 1:length(nphi))) - 1) / k
        }
        dim(nyt) <- dim(e2.tbl)

        nml <- recipient.ml (e2.alpha+mass*nyt, alph[j]*e2.tbl)
        if (runif(1)<exp(-ml+nml))
        { phi <- nphi
          yt <- nyt
          ml <- nml
        }
        else
        { rej[j] <- rej[j]+1
        }
      }

      # Update the log of the weight for this run.

      w[i] <- w[i] + recipient.ml (e2.alpha+mass*yt, alph[j+1]*e2.tbl) -
                     recipient.ml (e2.alpha+mass*yt, alph[j]*e2.tbl)

      W[i,j] <- w[i]
    }

    cat("Annealing run ",i,", log weight ",round(w[i],2),"\n",sep="")
  }

  # Find final estimates.

  log.sum.exp.w  <- log.sum.exp(w)

  ml <- log.sum.exp.w - log(N)

  se <- exp (log.sum.exp(log.sqdiff.exp(w,ml))/2-log(N) - ml)

  ess <- N / (1+var(exp(w-log.sum.exp.w+log(N))))

  rej <- rej/N

  cpu <- proc.time()[1] - stime

  return (ml, se, ess, cpu, rej, w, W)
}


# COMPUTE MARGINAL LIKELIHOOD WITH DONOR-WEIGHTED TRANSFER, BY ANN IMP SAMP, V2.
#
# The first three arguments are the procedure that generates parts of the
# donor model from its prior, the procedure that combines these parts into
# overall probabilties, and the procedure that calculates the marginal 
# likelihood for the recipient model.  The remaining arguments are as follows:
#
#    k          number of imaginary data points (default is infinite)
#    mass       mass of imaginary data (default k, required if k infinite)
#    N          number of annealing runs to do
#    alpha      Vector of powers to raise the likelihood to; 0 is added at the
#               front and 1 is added at the back
#
# The result is a list with the following elements:
#
#    ml     estimate of the log of the marginal likelihood for the entire data
#    se     standard error for the estimate of the log marginal likelihood
#    ess    effective sample size
#    cpu    Amount of user CPU time used for the computation
#
# The effective sample size is N divided by one plus the variance of the
# normalized importance weights.  It is a rough measure of the number of 
# unweighted data points that would be equivalent to the weighted sample.
#
# This version represents only phi and y, marginalizing away theta, and updates
# portions of phi and y by Metropolis-Hastings updates, with proposals to 
# modify parts of phi coming from the donor model's prior.

graph.dw.ais2 <- function (donor.gen.part, donor.comb.pr, recipient.ml, 
                           k=Inf, mass=k, N=100, alpha)
{
  stime <- proc.time()[1]

  # MAIN LOOP.  Does one annealing run each time around, storing the logs of
  # the importance weights for the runs in w. 

  w <- rep(0,N)

  alph <- c(0,alpha,1)

  W <- matrix (0, N, length(alpha)+1)
  rej <- matrix (0, length(alpha)+1, 4+e2.keep.cp)

  for (i in 1:N)
  {
    for (j in 1:(length(alpha)+1))
    {
      # Draw an independent state at the start of the run. 

      if (j==1)
      { phi <- donor.gen.part(e2.alpha)
        p <- donor.comb.pr (phi$sex, phi$ses, phi$pe, phi$iq, phi$cp)
        if (k==Inf)
        { yt <- p
        }
        else
        { y <- sample (length(p), k, replace=T, prob=p)
          yt <- (table (c(y, 1:length(p))) - 1) / k
        }
        dim(yt) <- dim(e2.tbl)
      }

      # Do M-H updates for each part after the start of the run.

      if (j>1)
      {
        ml <- recipient.ml (e2.alpha+mass*yt, alph[j]*e2.tbl)

        # sex

        nphi <- donor.gen.part(e2.alpha,1,phi$sex,phi$ses,phi$pe,phi$iq,phi$cp)
        np <- donor.comb.pr (nphi$sex, nphi$ses, nphi$pe, nphi$iq, nphi$cp)  
        if (k==Inf)
        { nyt <- np
        }
        else
        { ny <- sample (length(np), k, replace=T, prob=np)
          nyt <- (table (c(ny, 1:length(np))) - 1) / k
        }
        dim(nyt) <- dim(e2.tbl)

        nml <- recipient.ml (e2.alpha+mass*nyt, alph[j]*e2.tbl)
        if (runif(1)<exp(-ml+nml))
        { phi <- nphi
          p <- np
          yt <- nyt
          ml <- nml
        }
        else
        { rej[j,1] <- rej[j,1]+1
        }

        # ses

        nphi <- donor.gen.part(e2.alpha,2,phi$sex,phi$ses,phi$pe,phi$iq,phi$cp)
        np <- donor.comb.pr (nphi$sex, nphi$ses, nphi$pe, nphi$iq, nphi$cp)  
        if (k==Inf)
        { nyt <- np
        }
        else
        { ny <- sample (length(np), k, replace=T, prob=np)
          nyt <- (table (c(ny, 1:length(np))) - 1) / k
        }
        dim(nyt) <- dim(e2.tbl)

        nml <- recipient.ml (e2.alpha+mass*nyt, alph[j]*e2.tbl)
        if (runif(1)<exp(-ml+nml))
        { phi <- nphi
          p <- np
          yt <- nyt
          ml <- nml
        }
        else
        { rej[j,2] <- rej[j,2]+1
        }

        # pe

        nphi <- donor.gen.part(e2.alpha,3,phi$sex,phi$ses,phi$pe,phi$iq,phi$cp)
        np <- donor.comb.pr (nphi$sex, nphi$ses, nphi$pe, nphi$iq, nphi$cp)  
        if (k==Inf)
        { nyt <- np
        }
        else
        { ny <- sample (length(np), k, replace=T, prob=np)
          nyt <- (table (c(ny, 1:length(np))) - 1) / k
        }
        dim(nyt) <- dim(e2.tbl)

        nml <- recipient.ml (e2.alpha+mass*nyt, alph[j]*e2.tbl)
        if (runif(1)<exp(-ml+nml))
        { phi <- nphi
          p <- np
          yt <- nyt
          ml <- nml
        }
        else
        { rej[j,3] <- rej[j,3]+1
        }

        # iq

        nphi <- donor.gen.part(e2.alpha,4,phi$sex,phi$ses,phi$pe,phi$iq,phi$cp)
        np <- donor.comb.pr (nphi$sex, nphi$ses, nphi$pe, nphi$iq, nphi$cp)  
        if (k==Inf)
        { nyt <- np
        }
        else
        { ny <- sample (length(np), k, replace=T, prob=np)
          nyt <- (table (c(ny, 1:length(np))) - 1) / k
        }
        dim(nyt) <- dim(e2.tbl)

        nml <- recipient.ml (e2.alpha+mass*nyt, alph[j]*e2.tbl)
        if (runif(1)<exp(-ml+nml))
        { phi <- nphi
          p <- np
          yt <- nyt
          ml <- nml
        }
        else
        { rej[j,4] <- rej[j,4]+1
        }

        # cp

        if (e2.keep.cp)
        { nphi <- donor.gen.part(e2.alpha,5,
                                 phi$sex,phi$ses,phi$pe,phi$iq,phi$cp)
          np <- donor.comb.pr (nphi$sex, nphi$ses, nphi$pe, nphi$iq, nphi$cp)  
          if (k==Inf)
          { nyt <- np
          }
          else
          { ny <- sample (length(np), k, replace=T, prob=np)
            nyt <- (table (c(ny, 1:length(np))) - 1) / k
          }
          dim(nyt) <- dim(e2.tbl)

          nml <- recipient.ml (e2.alpha+mass*nyt, alph[j]*e2.tbl)
          if (runif(1)<exp(-ml+nml))
          { phi <- nphi
            p <- np
            yt <- nyt
            ml <- nml
          }
          else
          { rej[j,5] <- rej[j,5]+1
          }
        }
      }

      # Update the log of the weight for this run.

      w[i] <- w[i] + recipient.ml (e2.alpha+mass*yt, alph[j+1]*e2.tbl) -
                     recipient.ml (e2.alpha+mass*yt, alph[j]*e2.tbl)

      W[i,j] <- w[i]
    }

    cat("Annealing run ",i,", log weight ",round(w[i],2),"\n",sep="")
  }

  # Find final estimates.

  log.sum.exp.w  <- log.sum.exp(w)

  ml <- log.sum.exp.w - log(N)

  se <- exp (log.sum.exp(log.sqdiff.exp(w,ml))/2-log(N) - ml)

  ess <- N / (1+var(exp(w-log.sum.exp.w+log(N))))

  rej <- rej/N

  cpu <- proc.time()[1] - stime

  return (ml, se, ess, cpu, rej, w, W)
}
