# EXAMPLES OF PRIOR INFORMATION TRANSFER FOR TWO-WAY CONTINGENCY TABLES - PLOTS.
#
# Written by Radford M. Neal, July 2001.
#
# Produces the plots for the paper displaying the results of cont-ex.r.


# PLOTTING FUNCTIONS.  These are used to plot marginal likelihood profiles,
# which give the marginal likelihood for successively larger data sets, which
# are increasing prefixes of one data set.  The "skip" argument says how many 
# data points to skip to get the next data set (a value of 1 for skip means
# to plot the marginal likelihood for every size data set up to the total).
# The plot subtracts the profile under the true distribution, which is passed 
# as the "proft" parameter, so the values plotted are typcially negative,
# giving the penalty for using the model in question rather than than the
# true distribution.  (But this is not done if proft is NULL.)  The second 
# function adds labels, and should be called with the same arguments as the 
# first, apart from the initial argument. The profiles passed as arguments 
# may be interspersed with new values for the line type.

e1.plot.prof <- function (skip, proft, ...)
{
  upper <- lower <- 0

  a <- pairlist(...)

  while (!is.null(a))
  { if (length(a[[1]])!=1)
    { if (is.null(proft))
      { upper <- max(upper,a[[1]])
        lower <- min(lower,a[[1]])
      }
      else
      { upper <- max(upper,a[[1]]-proft)
        lower <- min(lower,a[[1]]-proft)
      }
    }
    a[[1]] <- NULL
  }

  if (is.null(proft))
  { plot (c(0,1.02*(length(pairlist(...)[[1]])-1)), c(lower,upper), type="n", 
          xlab="Number of observations", 
          ylab="Log marginal likelihood")
  }
  else
  { plot (c(0,1.02*(length(proft)-1)), c(lower,upper), type="n", 
          xlab="Number of observations", 
          ylab="Log marginal likelihood difference")
  }

  abline(0,0)

  a <- pairlist(...)
  type <- 1

  while (!is.null(a))
  { if (length(a[[1]])==1)
    { type <- a[[1]]
    }
    else if (is.null(proft))
    { lines (seq(0,(length(a[[1]])-1),skip), 
             a[[1]][seq(1,length(a[[1]]),skip)],lty=type)
    }
    else
    { lines (seq(0,(length(a[[1]])-1),skip), 
             (a[[1]]-proft)[seq(1,length(a[[1]]),skip)],lty=type)
    }
    a[[1]] <- NULL
  }
}

e1.plot.labels <- function (l, proft, ...)
{ 
  x <- length(proft)-1
  a <- pairlist(...)

  while (!is.null(a))
  { if (length(a[[1]])!=1)
    { if (is.null(proft))
      { text (x, a[[1]][x], l[[1]], cex=0.65)
      }
      else
      { text (x, a[[1]][x]-proft[x], l[[1]], cex=0.65)
      }
      l[[1]] <- NULL
    }
    a[[1]] <- NULL
  }
}


# MAKE THE PLOTS FOR THE PAPER.

ps.options(paper="special",width=2.9,height=2.9,pointsize=7,horizontal=F)
mypar <- function () par(mar=c(3,3,0.2,0.2),mgp=c(2,0.5,0))

postscript("fig-mlprof1a.ps"); mypar()
e1.plot.prof (10, NULL, e1.profd, e1.profr, e1.profj, e1.proft)
dev.off()

postscript("fig-mlprof1b.ps"); mypar()
e1.plot.prof (10, e1.proft, e1.profd, e1.profr, e1.profj)
e1.plot.labels (list("     d","     r","     s"), 
                e1.proft, e1.profd, e1.profr, e1.profj)
dev.off()

postscript("fig-mlprof2a.ps"); mypar()
e1.plot.prof (10, e1.proft, 
  e1.dw.fin.50$prof, e1.dw.fin.100$prof, e1.dw.fin.200$prof, e1.dw.fin.400$prof,
  e1.dw.fin.800$prof,
  3, e1.profd, e1.profr, e1.profj)
e1.plot.labels (
  list("       50","       100","       200","       400","       800",
       "     d","     r","     s"), 
  e1.proft,
  e1.dw.fin.50$prof, e1.dw.fin.100$prof, e1.dw.fin.200$prof, e1.dw.fin.400$prof,
  e1.dw.fin.800$prof,
  e1.profd, e1.profr, e1.profj)
dev.off()

postscript("fig-mlprof2b.ps"); mypar()
e1.plot.prof (10, e1.proft, 
  e1.dw.fin.50$prof, e1.dw.fin.100.50$prof, e1.dw.inf.50$prof,
  3, e1.profd, e1.profr, e1.profj)
e1.plot.labels (
  list("       50","       100","       inf",
       "     d","     r","     s"), 
  e1.proft,
  e1.dw.fin.50$prof, e1.dw.fin.100.50$prof, e1.dw.inf.50$prof,
  e1.profd, e1.profr, e1.profj)
dev.off()

postscript("fig-mlprof2c.ps"); mypar()
e1.plot.prof (10, e1.proft, 
  e1.dw.inf.50$prof, e1.dw.inf.100$prof, e1.dw.inf.200$prof, e1.dw.inf.400$prof,
  e1.dw.inf.800$prof,
  3, e1.profd, e1.profr, e1.profj)
e1.plot.labels (
  list("       50","       100","       200","       400","       800",
       "     d","     r","     s"), 
  e1.proft,
  e1.dw.inf.50$prof, e1.dw.inf.100$prof, e1.dw.inf.200$prof, e1.dw.inf.400$prof,
  e1.dw.inf.800$prof,
  e1.profd, e1.profr, e1.profj)
dev.off()

postscript("fig-mlprof2d.ps"); mypar()
e1.plot.prof (10, e1.proft, 
  e1.dw.inf.exp$prof,
  3, e1.profd, e1.profr, e1.profj)
e1.plot.labels (
  list("      exp",
       "     d","     r","     s"), 
  e1.proft,
  e1.dw.inf.exp$prof,
  e1.profd, e1.profr, e1.profj)
dev.off()

postscript("fig-mlprof3a.ps"); mypar()
e1.plot.prof (10, e1.proft, 
  e1.dw.inf.exp.u$prof,
  3, e1.profd, e1.profr, e1.profj)
lines (seq(0,(length(e1.proft)-1),10), 
       (e1.profi-e1.proft)[seq(1,length(e1.proft),10)])
e1.plot.labels (
  list("      ",
       "     d","     r","     s"), 
  e1.proft,
  e1.dw.inf.exp.u$prof,
  e1.profd, e1.profr, e1.profj)
dev.off()

postscript("fig-mlprof3b.ps"); mypar()
plot(c(0,0.98),c(-3620,-3537),type="n",
  xlab="quantiles of exp(500) distribution",
  ylab="log marginal likelihood")
points(seq(0,0.98,0.02),e1.dw.inf.exp$mlm,pch=1)
points(seq(0,0.98,0.02),e1.dw.inf.exp.u$mlm,pch="x")
e1.qpts <- c(0,50,100,200,300,400,500,1000)
text(pexp(e1.qpts/500),rep(-3537,10),as.character(e1.qpts))
dev.off()
