# EXAMPLES OF PRIOR INFORMATION TRANSFER FOR TWO-WAY CONTINGENCY TABLES.
#
# Written by Radford M. Neal, July 2001.
#
# This is Example 1 in my paper on "Transferring prior information between
# models using imaginary data".  All the variables are prefixed by "e1",
# for "example 1".  Some values are rounded to 4 decimal places so that
# tables won't be unnecessarily cumbersome.
#
# The log-exp.r and cont.r files must be executed before this one.
#
# See cont-ex-ais.r for a continuation that tries out use of annealed 
# importance sampling, and cont-ex-plt.r for the plots used in the paper.


# SPECIFICATIONS FOR THE EXAMPLE.

# Dimensions of the contingency table.

e1.n <- 7
e1.m <- 7

# Amount of data.

e1.sample.size <- 1000

# Parameters of the Dirichlet priors for the donor independence model.
# The two priors have the same sum, which is necessary for them to be
# consistent with the actual prior for joint probabilities (e1.jpr) below.

e1.dpr1 <- c(12,18,18,18,36,42,48)
e1.dpr2 <- c(12,12,12,24,24,54,54)

if (sum(e1.dpr1)!=sum(e1.dpr2)) stop("e1.dpr1 and d1.dpr2 don't have same sum")

# The Dirichlet distribution from which the actual parameters were drawn.
# It is constructed so that the two marginal distributions come from the 
# priors used in the donor independence model.

e1.jpr <- round (e1.dpr1 %o% e1.dpr2 / sum(e1.dpr1), 4)

# The Dirichlet(1,..,1) - ie, uniform - pro forma prior for the unrestricted 
# recipient model.

e1.rpr <- matrix(1,e1.n,e1.m)


# GENERATE THE PARAMETERS OF THE ACTUAL DISTRIBUTION AND THE ACTUAL DATA.
#
# The true parametes are in e1.true; the data is in e1.data1.

set.seed(1)

e1.true <- rgamma (e1.n*e1.m, as.vector(e1.jpr))
e1.true <- matrix (round (e1.true / sum(e1.true), 4), e1.n, e1.m)

e1.data <- cont.gen2 (e1.true, e1.sample.size)


# FIND THE LIKELIHOOD PROFILE OF THE DATA UNDER THE TRUE DISTRIBUTION.
# 
# Stored in e1.proft, and used as a reference profile when plotting.

e1.proft <- cont.profk2 (e1.data, e1.true)


# FIND MARGINAL LIKELIHOOD PROFILES FOR MODELS WITH DIRICHLET PRIORS.
#
# Finds marginal likelihood profiles using the distribution from which the
# parameters were actually drawn (j), the uniform pro forma prior of the 
# recipient model (r), the independence model with uniform priors (i),
# and the non-uniform prior for the donor model (d).

e1.profj <- cont.prof2 (e1.data, e1.jpr)
e1.profr <- cont.prof2 (e1.data, e1.rpr)
e1.profi <- cont.profi (e1.data, rep(1,e1.n), rep(1,e1.m))
e1.profd <- cont.profi (e1.data, e1.dpr1, e1.dpr2)


# FIND ESTIMATES USING DONOR-WEIGHTED PRIOR INFORMATION TRANSFER.  

set.seed(11)

# Use donor-weighted prior information transfer with mass parameter fixed at
# 10, 50, 100, 200, 400, and 800, with infinite k, and with k and the mass both 
# fixed at these values.  Done with simple Monte Carlo.

e1.dw.inf.10 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=2000,mass=10)

e1.dw.inf.50 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=2000,mass=50)

e1.dw.inf.100 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=2000,mass=100)

e1.dw.inf.200 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=2000,mass=200)

e1.dw.inf.400 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=2000,mass=400)

e1.dw.inf.800 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=2000,mass=800)

e1.dw.fin.10 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=10000,k=10)

e1.dw.fin.50 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=10000,k=50)

e1.dw.fin.100 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=10000,k=100)

e1.dw.fin.200 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=10000,k=200)

e1.dw.fin.400 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=10000,k=400)

e1.dw.fin.800 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=10000,k=800)

# Try a donor-weighted prior with mass not equal to k, with k finite.

set.seed(91)

e1.dw.fin.100.50 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"donor",N=10000,k=100,mass=50)

# Use donor-weighted prior information transfer with the prior for the mass
# parameter being exponential with mean 500.  Done with simple Monte Carlo.

set.seed(92)

e1.dw.inf.exp <- cont.pit (cont.pit.smc, e1.data, e1.rpr, e1.dpr1, e1.dpr2,
                            "donor", N=200, mass=500*qexp(seq(0,0.98,0.02)))


# FIND ESTIMATES USING JOINTLY-WEIGHTED PRIOR INFORMATION TRANSFER.  

set.seed(12)

e1.jw.inf.10 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"jointly",N=50000,mass=10)

e1.jw.inf.20 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"jointly",N=50000,mass=20)

e1.jw.inf.50 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"jointly",N=50000,mass=50)

set.seed(77)

e1.jw.inf.100 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"jointly",N=50000,mass=100)

set.seed(13)

e1.jw.fin.10 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"jointly",N=50000,k=10)

e1.jw.fin.20 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"jointly",N=50000,k=20)

e1.jw.fin.50 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"jointly",N=50000,k=50)

e1.jw.fin.50.20 <- 
  cont.pit.smc(e1.data,e1.rpr,e1.dpr1,e1.dpr2,"jointly",N=50000,k=50,mass=10)


# TRY TRANSFER FROM A DONOR INDEPENDENCE MODEL WITH UNIFORM PRIOR.
#
# Uses donor-weighted prior information transfer with the prior for the mass
# parameter being exponential with mean 500, but with the donor model having
# a uniform prior for the marginal probabilities, rather than the good prior.

set.seed(14)

e1.dw.inf.exp.u <- 
  cont.pit (cont.pit.smc, e1.data, e1.rpr, rep(1,e1.n), rep(1,e1.m),
            "donor", N=2000, mass=500*qexp(seq(0,0.98,0.02)))
