# PRIOR INFORMATION TRANSFER FOR MODELS OF TWO-WAY CONTINGENCY TABLES.
#
# Written by Radford M. Neal, July 2001.
#
# These functions were used for Example 1 in my paper on ``Transferring
# prior information between models using imaginary data''.  
#
# These functions use the functions defined in log-exp.r.

# -----------------------------------------------------------------------------

# GENERATE ONE-WAY DATA FROM A GIVEN PROBABILITY TABLE.  The vector p
# gives the probabilities for each category; n is the sample size.  The
# result is a vector of integers from 1 to length(p).

cont.gen1 <- function (p, n)
{
  sample (length(p), n, replace=T, prob=p)
}


# GENERATE TWO-WAY DATA FROM A GIVEN PROBABILITY TABLE.  The matrix p
# gives the probabilities for each combination of categorical values;
# n is the sample size.  The result is a matrix with n rows and two
# columns, the first containing integers from 1 to nrow(p), the second
# containing integers from 1 to ncol(p).

cont.gen2 <- function (p, n)
{ 
  r <- matrix(0,n,2)

  s <- cont.gen1(as.vector(p),n)

  r[,1] <- (s-1) %% nrow(p) + 1
  r[,2] <- (s-1) %/% nrow(p) + 1

  r
}


# GENERATE TWO-WAY DATA FROM AN INDEPENDENCE MODEL.  The vectors p1 and
# p2 give the probabilities for the two categorical variables, assumed
# independent; n is the sample size.  The result is an n by 2 matrix.

cont.geni <- function (p1, p2, n)
{
  cbind (cont.gen1(p1,n), cont.gen1(p2,n))
}

# -----------------------------------------------------------------------------

# PRODUCE LIKELIHOOD PROFILE OF ONE-WAY DATA UNDER A KNOWN DISTRIBUTION.
# Returns a vector with length(d)+1 elements, containing the log probabilities 
# of obtaining the first i observed data items in d, for i ranging from
# 0 to length(d).  (Note that this log probability is always 0 when i is 0.)
# The probabiities are assumed known, as given by the vector p.

cont.profk1 <- function (d, p)
{
  n <- length(d)
  m <- rep(0,n+1)
  
  for (i in 1:n)
  { m[i+1] <- m[i] + log(p[d[i]])
  }

  m
}


# PRODUCE LIKELIHOOD PROFILE OF TWO-WAY DATA UNDER A KNOWN DISTRIBUTION.
# Returns a vector with nrow(d)+1 elements, containing the log probabilities 
# of obtaining the first i observed data items (rows of d), for i ranging 
# from 0 to nrow(d).  (Note that this log probability is always 0 when i is 0.)
# The probabiities are assumed known, as given by the matrix p.

cont.profk2 <- function (d, p)
{
  cont.profk1 (nrow(p)*(d[,2]-1) + d[,1], as.vector(p))
}


# PRODUCE LIKELIHOOD PROFILE OF TWO-WAY DATA UNDER A KNOWN INDEPENDENCE MODEL.
# Returns a vector with nrow(d)+1 elements, containing the log probabilities 
# of obtaining the first i observed data items (rows of d), for i ranging 
# from 0 to nrow(d).  (Note that this log probability is always 0 when i is 0.)
# The two categorical variables are assumed to be independent, with known
# probabilities given by the vectors p1 and p2.

cont.profki <- function (d, p1, p2)
{ 
  cont.profk1(d[,1],p1) + cont.profk1(d[,2],p2)
}

# -----------------------------------------------------------------------------

# FIND THE LOG MARGINAL LIKELIHOOD FOR ONE-WAY DATA WITH A DIRICHLET PRIOR.
# Computes the log of the marginal likelihood of the categorical data in d,
# under the Dirichlet prior with parameters given by the vector a.

cont.marg1 <- function (d, a)
{
  n <- length(d)
  c <- table(c(d,1:length(a))) - 1
  A <- sum(a)

  sum (lgamma(a+c)-lgamma(a)) - (lgamma(A+n)-lgamma(A))
}


# FIND THE LOG MARGINAL LIKELIHOOD FOR TWO-WAY DATA WITH A DIRICHLET PRIOR.
# Computes the log of the marginal likelihood of the two-way data in d,
# under the Dirichlet prior with parameters given by the matrix a.

cont.marg2 <- function (d, a)
{
  cont.marg1 (nrow(a)*(d[,2]-1) + d[,1], as.vector(a))
}


# FIND LOG MARGINAL LIKELIHOOD FOR TWO-WAY DATA UNDER AN INDEPENDENCE MODEL.
# Computes the log of the marginal likelihood of the two-way data in d,
# under a model assuming independence, with Dirichlet priors given by the 
# vectors a1 and a2.

cont.margi <- function (d, a1, a2)
{
  cont.marg1(d[,1],a1) + cont.marg1(d[,2],a2)
}

# -----------------------------------------------------------------------------

# PRODUCE MARGINAL LIKELIHOOD PROFILE FOR ONE-WAY DATA WITH DIRICHLET PRIOR.
# Returns a vector with length(d)+1 elements, containing the log marginal
# likelihoods for the first i observed data items in d, for i ranging
# from 0 to length(d).  (Note that the log marginal likelihood when i is 0 is
# always 0).  The prior is assumed to be Dirichlet with parameter vector a.
#
# Note that the results of this procedure should match those of cont.marg1,
# applied to the appropriate portions of d.

cont.prof1 <- function (d, a)
{
  n <- length(d)
  A <- sum(a)
  m <- rep(0,n+1)
  u <- a

  for (i in 1:n)
  { m[i+1] <- m[i] + log(u[d[i]]/(A+i-1))
    u[d[i]] <- u[d[i]] + 1
  }

  if (abs(m[n+1]-cont.marg1(d,a))>1e-5)
  { stop("Something's wrong: answer doesn't match that from cont.marg1")
  }

  m
}


# PRODUCE MARGINAL LIKELIHOOD PROFILE FOR TWO-WAY DATA WITH DIRICHLET PRIOR.
# Returns a vector with nrow(d)+1 elements, containing the log marginal
# likelihoods for the first i observed data items (rows of d), for i ranging
# from 0 to nrow(d).  (Note that the log marginal likelihood when i is 0 is
# always 0).  The prior is assumed to be Dirichlet with parameters given by
# the matrix a.
#
# Note that the results of this procedure should match those of cont.marg2,
# applied to the appropriate portions of d.

cont.prof2 <- function (d, a)
{
  cont.prof1 (nrow(a)*(d[,2]-1) + d[,1], as.vector(a))
}


# PRODUCE MARGINAL LIKELIHOOD PROFILE FOR TWO-WAY DATA WITH INDEPENDENCE MODEL.
# Returns a vector with nrow(d)+1 elements, containing the log marginal
# likelihoods for the first i observed data items (rows of d), for i ranging
# from 0 to nrow(d).  (Note that the log marginal likelihood when i is 0 is
# always 0).  The two categorical variables are assumed to be independent,
# with their probabilities having Dirichlet priors with parameters given by
# the vectors a1 and a2.
#
# Note that the results of this procedure should match those of cont.margi,
# applied to the appropriate portions of d.

cont.profi <- function (d, a1, a2)
{
  cont.prof1(d[,1],a1) + cont.prof1(d[,2],a2)
}

# -----------------------------------------------------------------------------

# MODEL TWO-WAY DATA USING PRIOR TRANSFER WITH SIMPLE MONTE CARLO.
# 
# Estimates the marginal likelihood profile and the posterior probabilities
# for two-way data modeled by an unrestricted model that uses a prior obtained
# by prior information transfer from an independence model.  The pro forma
# prior for the unrestricted model is Dirichlet, as are the priors for the
# marginal probabilities in the donor model (which are a priori independent).  
# Either donor-weighted or jointly-weighted prior information transfer may
# be used.  Computations are done using simple Monte Carlo sampling of
# imaginary data (or of donor model parameters, if the amount of imaginary
# data is infinite).
#
# This function uses a fixed mass parameter.  Results with a variable mass 
# parameter can be obtained using cont.pit.
#
# The arguments are as follows:
#
#    d          the n by 2 data matrix
#    ar         matrix of parameters of the pro forma Dirichlet prior
#    ad1, ad2   parameter vectors for the Dirichlet priors of the donor model
#    wt         Type of weighting, either "donor" or "jointly"
#    k          number of imaginary data points (default is infinite)
#    mass       mass of imaginary data (default k, no default if k is infinite)
#    N          sample size for Monte Carlo approximation (default 100)
#
# The data in the first column of d should be integers from 1 to nrow(ar);
# the data in the second column should be integers from 1 to ncol(ar).  The
# length of ad1 should equal nrow(ar); the length of ad2 should equal ncol(ar).
#
# The result is a list with the following elements:
#
#    pp     matrix of estimated posterior mean probabilities
#    ml     estimate of the log of the marginal likelihood for the entire data
#    se     standard error for the estimate of the log marginal likelihood
#    pn     log normalizing constant for the prior, 0 for donor-weighted prior
#    prof   vector of nrow(d)+1 estimates of the log marginal likelihoods for
#           the first i data items, with i ranging from 0 to nrow(d)
#    cpu    Amount of user CPU time used for the computation

# Notes on the implementation:  To avoid floating-point underflow problems, 
# the quantities that are accumulated in order to compute the estimates are 
# represented by logs.  The estimate and standard error for the log marginal 
# likelihood is found by assuming that the standard error is small enough that 
# the log function is locally linear.

cont.pit.smc <- function (d, ar, ad1, ad2, wt, k=Inf, mass=k, N=100)
{
  stime <- proc.time()[1]

  # Find some dimensions.

  d1 <- nrow(ar)
  d2 <- ncol(ar)

  n <- nrow(d)

  # Check arguments for validity.

  if (length(ad1)!=d1 || length(ad2)!=d2)
  { stop(
     "Prior parameters for donor and recipient models have inconsistent sizes")
  }

  if (mass==Inf)
  { stop("The mass argument must be given a finite value")
  }

  if (wt!="donor" && wt!="jointly")
  { stop("Weight type (wt) must be either \"donor\" or \"jointly\"")
  }
  jointly.weighted <- wt=="jointly"

  # Make two-way contingency table containing counts from the data.

  dt <- matrix (table (c(d1*(d[,2]-1) + d[,1], 1:(d1*d2))) - 1, d1, d2)

  # MAIN LOOP.  Accumulates the data on which estimates will be based in 
  # pp, mlv, and prof.  

  prof <- rep(-Inf,n+1)      # logs of sums for finding marg. likelihood profile
  pp   <- matrix(-Inf,d1,d2) # logs of sums for estimating probabilities
  mlv  <- rep(0,N)           # logs of probabilities of data for N values of y
  w    <- rep(0,N)           # logs of weights to be given to each value of y,
                             #   all zero for donor-weighted transfer
  if (jointly.weighted)
  { sum.ar <- sum(ar)
  }

  for (i in 1:N)
  { 
    # Generate parameters for the donor model from its prior.

    phi1 <- rgamma(d1,ad1); phi1 <- phi1/sum(phi1)
    phi2 <- rgamma(d2,ad2); phi2 <- phi2/sum(phi2)
    p <- phi1 %o% phi2

    # Generate a table of counts from imaginary data, adjusted to have the
    # desired total mass; use probabilities if the amount of imaginary
    # data is infinite.

    if (k==Inf)
    { yc <- mass * p
    }
    else
    { y <- sample (d1*d2, k, replace=T, prob=as.vector(p))
      yt <- table (c(y, 1:(d1*d2))) - 1
      yc <- (mass/k) * matrix (yt, d1, d2)
    }

    # Find the log marginal likelihood profile using the pro forma prior 
    # adjusted by the imaginary data.  

    prf <- cont.prof2 (d, ar + yc)

    # Find the posterior mean of the probabilities conditional on the present
    # imaginary data.

    pm <- dt+ar+yc
    pm <- pm/sum(pm)

    # Find the log weight for this y when using jointly-weighted transfer.  
    # (The log weight is left at zero for donor-weighted transfer.)

    if (jointly.weighted)
    { w[i] <- sum(lgamma(ar+yc)-lgamma(ar)) - 
                (lgamma(sum.ar+mass)-lgamma(sum.ar))
    }

    # Update the variables accumulating statistics for use in finding estimates.

    mlv[i] <- prf[n+1]+w[i]
    prof <- log.add.exp (prof, prf+w[i])
    pp <- log.add.exp (pp, log(pm)+mlv[i])
  }

  # Compute and return the final estimates.

  log.sum.exp.w <- log.sum.exp(w)
  log.sum.exp.mlv <- log.sum.exp(mlv)

  ml <- log.sum.exp.mlv - log.sum.exp.w

  if (mass==0) 
  { se <- 0
  } 
  else
  { se <- exp (log.sum.exp(log.sqdiff.exp(mlv,ml+w))/2 - log.sum.exp.w - ml)
  }

  prof <- prof - log.sum.exp.w

  if (abs(ml-prof[n+1])>1e-5) 
  { stop("Something's wrong: ml and prof[n+1] don't match")
  }

  pp <- exp(pp-log.sum.exp.mlv)

  if (abs(1-sum(pp))>1e-5)
  { stop("Something's wrong: pp doesn't sum to one")
  }

  pn <- log.sum.exp.w - log(N)

  cpu <- proc.time()[1] - stime

  return (pp, ml, se, pn, prof, cpu)
}

# -----------------------------------------------------------------------------

# MODEL TWO-WAY DATA USING PRIOR TRANSFER WITH ANNEALED IMPORTANCE SAMPLING.
# 
# Estimates the marginal likelihood profile and the posterior probabilities
# for two-way data modeled by an unrestricted model that uses a prior obtained
# by prior information transfer from an independence model.  The pro forma
# prior for the unrestricted model is Dirichlet, as are the priors for the
# marginal probabilities in the donor model (which are a priori independent).  
# Either donor-weighted or jointly-weighted prior information transfer may
# be used.  Computations are done using annealed importance sampling, with
# one Gibbs sampling scan done at each stage of annealing.
#
# This function uses a fixed mass parameter.  Results with a variable mass 
# parameter can be obtained using cont.pit.
#
# The arguments are as follows:
#
#    d          the n by 2 data matrix
#    ar         matrix of parameters of the pro forma Dirichlet prior
#    ad1, ad2   parameter vectors for the Dirichlet priors of the donor model
#    wt         Type of weighting, either "donor" or "jointly"
#    k          number of imaginary data points (default is infinite)
#    mass       mass of imaginary data (default k, required when k is infinite)
#    N          number of annealing runs to do
#    alpha      Vector of powers to raise the likelihood to; 0 added at front
#    omega      Vector of powers to raise the imaginary likelihood to.  Must
#               start with 0 if the prior is jointly-weigthed; element equal to
#               mass is added to the end.  Defaults to nothing, which is 
#               appropriate for a donor-weighted prior
#
# The data in the first column of d should be integers from 1 to nrow(ar);
# the data in the second column should be integers from 1 to ncol(ar).  The
# length of ad1 should equal nrow(ar); the length of ad2 should equal ncol(ar).
#
# The result is a list with the following elements:
#
#    pp     matrix of estimated posterior mean probabilities
#    ml     estimate of the log of the marginal likelihood for the entire data
#    se     standard error for the estimate of the log marginal likelihood
#    pn     log normalizing constant for the prior, 0 for donor-weighted prior
#    ess    effective sample size
#    cpu    Amount of user CPU time used for the computation
#    w, w0, W  Intermediate importance weights for use in debugging and tuning
#
# The effective sample size is N divided by one plus the variance of the
# normalized importance weights.  It is a rough measure of the number of 
# unweighted data points that would be equivalent to the weighted sample.
#
# Donor-weighted priors with infinite k are not currently implemented.

# Notes on the implementation:  To avoid floating-point underflow problems, 
# the quantities that are accumulated in order to compute the estimates are 
# represented by logs.  The estimate and standard error for the log marginal 
# likelihood is found by assuming that the standard error is small enough that 
# the log function is locally linear.  When jointly-weighted transfer is done,
# the standard error for the log marginal likelihood is computed ignoring the
# error in the estimate of the log normalizing constant for the prior.  This
# probably leads to an overstatement of the error, since the two quantities
# are subtracted, and are likely to have errors that are positively correlated.

cont.pit.ais <- 
  function (d, ar, ad1, ad2, wt, k=Inf, mass=k, N=100, alpha, omega=numeric(0))
{
  stime <- proc.time()[1]

  # Find some dimensions.

  d1 <- nrow(ar)
  d2 <- ncol(ar)

  n <- nrow(d)

  # Check arguments for validity.

  if (length(ad1)!=d1 || length(ad2)!=d2)
  { stop(
     "Prior parameters for donor and recipient models have inconsistent sizes")
  }

  if (mass==Inf)
  { stop("The mass argument must be given a finite value")
  }

  if (wt!="donor" && wt!="jointly")
  { stop("Weight type (wt) must be either \"donor\" or \"jointly\"")
  }
  jointly.weighted <- wt=="jointly"

  if (jointly.weighted && (length(omega)==0 || omega[1]!=0))
  { stop(
"Annealing schedule must start with omega=0 when using a jointly-weigthed prior"
    )
  }

  # Make two-way contingency table containing counts from the data.

  dt.vec <- table (c(d1*(d[,2]-1) + d[,1], 1:(d1*d2))) - 1

  # MAIN LOOP.  Does one annealing run each time around, storing the logs of
  # the importance weights for the runs in w.  The importance weights at
  # the point where we reach the prior (omega equal to mass) are recorded
  # in w0 (this is needed only for jointly-weighted transfer).  The logs
  # of the weighted sums of theta values are stored in pp.

  ar.vec <- as.vector(ar)

  w <- rep(0,N)
  w0 <- rep(0,N)

  W <- matrix (0, N, length(omega)+length(alpha)+1)

  pp <- rep(-Inf,d1*d2)

  omga <- c(omega,rep(mass,length(alpha)+2))
  alph <- c(rep(0,length(omega)+1),alpha,1)

  for (i in 1:N)
  {
    for (j in 1:(length(omega)+length(alpha)+1))
    {
      # Draw an independent state at the start of the run. 

      if (j==1)
      { phi1 <- rgamma(d1,ad1); phi1 <- phi1/sum(phi1)
        phi2 <- rgamma(d2,ad2); phi2 <- phi2/sum(phi2)
        p <- as.vector (phi1 %o% phi2)
        if (k==Inf)
        { yt <- p
        }
        else
        { y <- sample (d1*d2, k, replace=T, prob=p)
          yt <- (table (c(y, 1:(d1*d2))) - 1) / k
        }
        theta <- rgamma (d1*d2, ar.vec + omga[j]*yt)
        theta <- theta/sum(theta)
      }

      # Do a Gibbs sampling scan after the start of the run. 

      if (j>1)
      {
        # Sample for theta given data and y.

        theta <- rgamma (d1*d2, ar.vec+omga[j]*yt+alph[j]*dt.vec) 
        theta <- theta/sum(theta)

        # Sample for y given phi and theta.  Skip when k is infinite.

        if (k!=Inf)
        {
          p <- as.vector (phi1 %o% phi2) * theta^(omga[j]/k)
          p <- p/sum(p)

          if (jointly.weighted)
          { y <- sample (d1*d2, k, replace=T, prob=p)
            yt <- (table (c(y, 1:(d1*d2))) - 1) / k
          }
          else
          { for (h in 1:k)
            { yt[y[h]] <- yt[y[h]] - 1/k
              ph <- p * exp (
                lgamma(ar.vec+omga[j]*yt) - lgamma(ar.vec+omga[j]*(yt+1/k)))
              y[h] <- sample(d1*d2, 1, prob=ph/sum(ph))
              yt[y[h]] <- yt[y[h]] + 1/k
            }
          }
        }

        # Sample for phi given y when k is finite.  When k is infinite, sample
        # for phi and yt simultaneously, with a Metropolis-Hastings update 
        # that proposes a phi (and consequent yt) drawn from the donor's prior.

        if (k==Inf)
        { if (jointly.weighted)
          { nphi1 <- rgamma(d1,ad1); nphi1 <- nphi1/sum(nphi1)
            nphi2 <- rgamma(d2,ad2); nphi2 <- nphi2/sum(nphi2)
            nyt <- as.vector (nphi1 %o% nphi2)
            nyt <- nyt/sum(nyt)
            d <- sum(omga[j]*(yt-nyt)*log(theta))
            if (runif(1)<exp(-d))
            { phi1 <- nphi1
              phi2 <- nphi2
              yt <- nyt
            }
          }
          else
          { stop("Donor-weighted priors with infinite k not implemented yet")
          }
        }
        else
        { yt1 <- table (c((y-1) %%  d1 + 1, 1:d1)) - 1
          yt2 <- table (c((y-1) %/% d1 + 1, 1:d2)) - 1
          phi1 <- rgamma(d1,ad1+yt1); phi1 <- phi1/sum(phi1)
          phi2 <- rgamma(d2,ad2+yt2); phi2 <- phi2/sum(phi2)
        }
      }

      # Update the log of the weight for this run.

      w[i] <- w[i] + sum ((omga[j+1]*yt + alph[j+1]*dt.vec) * log(theta)) -
                     sum ((omga[j]*yt + alph[j]*dt.vec) * log(theta))

      if (!jointly.weighted && omga[j+1]!=omga[j])
      { w[i] <- w[i] + lgamma(sum(ar.vec)+omga[j+1]) -
                       lgamma(sum(ar.vec)+omga[j]) - 
                       sum (lgamma (ar.vec+omga[j+1]*yt)) + 
                       sum (lgamma (ar.vec+omga[j]*yt))
      }

      # Record the weight at the point where the prior is reached.  These
      # weights get left at zero for donor-weighted transfer.

      if (j==length(omega))
      { w0[i] <- w[i]
      }

      W[i,j] <- w[i]
    }

    pp <- log.add.exp (pp, log(theta)+w[i])

    cat("Annealing run ",i,", log weight ",round(w[i],2),"\n",sep="")
  }

  # Find final estimates.

  log.sum.exp.w  <- log.sum.exp(w)
  log.sum.exp.w0 <- log.sum.exp(w0)

  pn <- log.sum.exp.w0 - log(N)

  ml <- log.sum.exp.w - log(N)

  se <- exp (log.sum.exp(log.sqdiff.exp(w,ml))/2-log(N) - ml)

  ml <- ml - pn

  pp <- exp(pp-log.sum.exp.w)

  if (abs(1-sum(pp))>1e-5)
  { stop("Something's wrong: pp doesn't sum to one")
  }

  ess <- N / (1+var(exp(w-log.sum.exp.w+log(N))))

  cpu <- proc.time()[1] - stime

  return (pp, ml, pn, se, ess, cpu, w, w0, W)
}

# -----------------------------------------------------------------------------

# MODEL TWO-WAY DATA USING PRIOR TRANSFER WITH VARIABLE MASS PARAMETER.
# 
# This function implements prior information transfer with a variable mass of
# imaginary data, using one or the other of the cont.pit.smc or cont.pit.mcmc 
# functions defined above to do the computations for each possible mass value.
#
# The arguments and results of this function are the same as those of 
# cont.pit.smc or cont.pit.ais, whichever is specified as the first
# argument of this function, except that the mass argument can be a vector 
# (though it still defaults to the single value k), a vector of log marginal
# likelihoods for each element of this mass vector (called "mlm") and 
# associated standard errors (called "sem") are returned along with the overall 
# log marginal likelihood (called "ml"), and the log normalizing constant for
# the prior ("pn") is not returned.  The prior for the mass parameter is 
# uniform over the elements in the mass vector.  A continuous prior can be 
# approximated by passing a vector of equally-spaced quantiles as the mass 
# argument - for example, qexp(seq(0,0.99,0.01)) approximates an exponential 
# prior. 

# Notes on the implementation:  To avoid floating-point underflow problems, 
# the quantities that are accumulated in order to compute the estimates are 
# represented by logs.  The estimate and standard error for the log marginal 
# likelihood is found by assuming that the standard error is small enough that 
# the log function is locally linear.

cont.pit <- function (func, d, ar, ad1, ad2, wt, k=Inf, mass=k, ...)
{
  stime <- proc.time()[1]

  # Find some dimensions.

  d1 <- nrow(ar)
  d2 <- ncol(ar)

  n <- nrow(d)
  M <- length(mass)

  # MAIN LOOP.  Uses a different mass parameter each time around.

  mlm <- rep(0,M)            # logs of marg. likelihood estimates for each mass
  sem <- rep(0,M)            # standard errors for estimates in mlm
  prof <- rep(-Inf,n+1)      # logs of sums for finding marg. likelihood profile
  pp   <- matrix(-Inf,d1,d2) # logs of sums for estimating probabilities

  for (m in 1:M)
  { 
    r <- func(d,ar,ad1,ad2,wt,k,mass[m],...)

    mlm[m] <- r$ml
    sem[m] <- r$se

    if (!is.null(r$prof)) 
    { prof <- log.add.exp (prof, r$prof)
    }

    pp <- log.add.exp (pp, log(r$pp)+r$ml)
  }

  # Compute and return final estimates.

  log.sum.exp.mlm <- log.sum.exp(mlm)

  ml <- log.sum.exp.mlm - log(M)
  se <- exp (log.sum.exp(2*(log(sem)+mlm))/2 - log(M) - ml)

  if (!is.null(r$prof))
  {  
    prof <- prof - log(M)

    if (abs(ml-prof[n+1])>1e-5) 
    { stop("Something's wrong: ml and prof[n+1] don't match")
    }
  }

  pp <- exp(pp-log.sum.exp.mlm) 

  if (abs(1-sum(pp))>1e-5)
  { stop("Something's wrong: pp doesn't sum to one")
  }

  cpu <- proc.time()[1] - stime

  if (is.null(r$prof))
  { return (pp, ml, se, mlm, cpu)
  }
  else
  { return (pp, ml, se, prof, mlm, sem, cpu)
  }
}

# -----------------------------------------------------------------------------

# FIND NORMALIZATION OF JOINTLY-WEIGHTED PRIOR BY SAMPLING FROM RECIPIENT PRIOR.
#
# Only priors with mass equal to the number of imaginary data points are 
# handled.

cont.jw.norm.r <- function (ar, ad1, ad2, k, N)
{
  stime <- proc.time()[1]

  d1 <- nrow(ar)
  d2 <- ncol(ar)

  pnv <- rep(0,N)

  for (i in 1:N)
  {
    theta <- rgamma (d1*d2, ar)
    theta <- theta/sum(theta)
    dim(theta) <- dim(ar)

    y <- cont.gen2 (theta, k)

    pnv[i] <- cont.margi(y,ad1,ad2)
  }

  pn <- log.sum.exp(pnv) - log(N)
  se <- exp (log.sum.exp(log.sqdiff.exp(pnv,pn))/2-log(N) - pn)

  cpu <- proc.time()[1] - stime

  return (pn, se, pnv, cpu)
}


# FIND NORMALIZATION OF JOINTLY-WEIGHTED PRIOR BY SAMPLING FROM DONOR PRIOR.
#
# Only priors with mass equal to the number of imaginary data points are 
# handled.

cont.jw.norm.d <- function (ar, ad1, ad2, k, N)
{
  stime <- proc.time()[1]

  d1 <- nrow(ar)
  d2 <- ncol(ar)

  pnv <- rep(0,N)

  for (i in 1:N)
  {
    phi1 <- rgamma (d1, ad1)
    phi1 <- phi1/sum(phi1)
    phi2 <- rgamma (d2, ad2)
    phi2 <- phi2/sum(phi2)

    y <- cont.geni (phi1, phi2, k)

    pnv[i] <- cont.marg2(y,ar)
  }

  pn <- log.sum.exp(pnv) - log(N)
  se <- exp (log.sum.exp(log.sqdiff.exp(pnv,pn))/2-log(N) - pn)

  cpu <- proc.time()[1] - stime

  return (pn, se, pnv, cpu)
}


# FIND NORMALIZATION OF JOINTLY-WEIGHTED PRIOR BY ANNEALED IMPORTANCE SAMPLING.
#
# This function can handle a mass of imaginary data not equal to k.  An
# annealing schedule (omega) must be supplied, which must start with zero.

cont.jw.norm.ais <- function (ar, ad1, ad2, k, mass=k, N, omega)
{
  stime <- proc.time()[1]

  # Find some dimensions.

  d1 <- nrow(ar)
  d2 <- ncol(ar)

  # Check arguments for validity.

  if (length(ad1)!=d1 || length(ad2)!=d2)
  { stop(
     "Prior parameters for donor and recipient models have inconsistent sizes")
  }

  if (length(omega)==0 || omega[1]!=0)
  { stop("Annealing schedule must start with omega=0")
  }

  # MAIN LOOP.  Does one annealing run each time around.

  ar.vec <- as.vector(ar)

  w <- rep(0,N)

  W <- matrix (0, N, length(omega))

  omga <- c(omega,mass)

  for (i in 1:N)
  {
    for (j in 1:length(omega))
    {
      # Draw an independent state at the start of the run.  

      if (j==1)
      { phi1 <- rgamma(d1,ad1); phi1 <- phi1/sum(phi1)
        phi2 <- rgamma(d2,ad2); phi2 <- phi2/sum(phi2)
        p <- phi1 %o% phi2
        y <- sample (d1*d2, k, replace=T, prob=as.vector(p))
        yt <- table (c(y, 1:(d1*d2))) - 1
        theta <- rgamma (d1*d2, ar.vec)
        theta <- theta/sum(theta)
      }

      # Do a Gibbs sampling scan after the start of the run.  

      if (j>1)
      {
        # Sample for theta given data and y.

        theta <- rgamma (d1*d2, ar.vec+(omga[j]/k)*yt) 
        theta <- theta/sum(theta)

        # Sample for y given phi and theta.

        p <- as.vector (phi1 %o% phi2) * theta^(omga[j]/k)
        p <- p/sum(p)
        y <- sample (d1*d2, k, replace=T, prob=p)
        yt <- table (c(y, 1:(d1*d2))) - 1

        # Sample for phi given y.

        yt1 <- table (c((y-1) %%  d1 + 1, 1:d1)) - 1
        yt2 <- table (c((y-1) %/% d1 + 1, 1:d2)) - 1
        phi1 <- rgamma(d1,ad1+yt1); phi1 <- phi1/sum(phi1)
        phi2 <- rgamma(d2,ad2+yt2); phi2 <- phi2/sum(phi2)
      }

      # Update the log of the weight for this run.

      w[i] <- w[i] + sum ((yt*(omga[j+1]-omga[j])/k) * log(theta))

      W[i,j] <- w[i]
    }

    # cat("Annealing run ",i,", log weight ",round(w[i],2),"\n",sep="")
  }

  # Find final estimate and standard error.

  pn  <- log.sum.exp(w) - log(N)
  se <- exp (log.sum.exp(log.sqdiff.exp(w,pn))/2-log(N) - pn)

  cpu <- proc.time()[1] - stime

  return (pn, se, w, W, cpu)
}
