# INVESTIGATION OF PRIOR TRANSFER FOR BERNOULLI MODELS WITH UNIFORM PRIOR.
#
# Written by Radford M. Neal, July 2001.
#
# These functions investigate what happens when one uses various forms of
# prior information transfer between identical models for Bernoulli data,
# with uniform prior for the unknown probability.  Each function computes
# an approximation to the prior after transfer.
#
# Some results one can get with these functions (which can also be found 
# analytically) are mentioned in the section on idempotence in my paper
# on "Transferring prior information between models using imaginary data".


# GRID OF POINTS TO EVALUATE PRIOR DENSITY AT.

grid <- seq(0.01,0.99,0.01)


# PLOT DENSITY FROM POINTS ON GRID.

unif.plt <- function (d, xlim=c(0,1), ylim=c(0,max(d)))
{
  plot (grid, d, type="l", xlim=xlim, ylim=ylim)
}


# FIND PRIOR DENSITY RESULTING FROM DONOR-WEIGHTED TRANSFER.  Omega
# should not be infinite, which means it is required if k is left to default
# to infinity.

unif.dw <- function (k=Inf, omega=k, N=10000)
{
  if (omega==Inf)
  { stop("Infinite value for omega not allowed")
  }

  phi <- seq(0,1,length=N)

  if (k==Inf)
  { s <- phi
  }
  else
  { s <- rbinom(N,k,phi) / k
  }

  G <- 0
  for (i in 1:N)
  { G <- G + dbeta (grid, omega*s[i]+1, omega*(1-s[i])+1) / N
  }

  G
}


# FIND PRIOR DENSITY RESULTING FROM JOINTLY-WEIGHTED TRANSFER. Omega
# should not be infinite, which means it is required if k is left to default
# to infinity.

unif.jw <- function (k=Inf, omega=k, N=10000)
{
  if (omega==Inf)
  { stop("Infinite value for omega not allowed")
  }

  phi <- seq(0,1,length=N)

  if (k==Inf)
  { s <- phi
  }
  else
  { s <- rbinom(N,k,phi) / k
  }

  w <- lgamma(omega*s+1) + lgamma(omega*(1-s)+1) - lgamma(omega+2)
  w <- exp(w-max(w))
  w <- w/sum(w)

  G <- 0
  for (i in 1:N)
  { G <- G + w[i] * dbeta (grid, omega*s[i]+1, omega*(1-s[i])+1)
  }

  G
}
